unit DataClasses;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Controls, StdCtrls, ExtCtrls, Forms, dialogs, Graphics, xmlUtil;

type
  TDataClass = class;

  TDataItem = class(TPanel)
    private
      _Prev, _Next : TDataItem;
      _Parent      : TDataClass;
      function ReadTitle : String; inline;
      procedure WriteTitle(v : String); inline;
      procedure WriteParent(P : TDataClass);
      procedure RemoveClick(Sender : TObject);
      procedure DoClick(Sender : TObject);
    protected
      ButRem     : TButton;
      ButMod     : TButton;
      TitleEl    : TLabel;
    public
      constructor Create;
      procedure Remove;
      procedure Activate; virtual;
      procedure Deactivate; virtual;
      property Next : TDataItem read _Next;
      property Prev : TDataItem read _Prev;
      procedure InsertAfter(P : TDataItem);
      procedure InsertFirst(P : TDataClass);
      property Title : String read ReadTitle write WriteTitle;
      property Parent : TDataClass read _Parent write WriteParent;
      procedure Edit(Sender : TObject); virtual;
      function DataToXML : XML.TElement; virtual; abstract;
      procedure XMLToData(Data : XML.TElement); virtual; abstract;
      destructor Destroy; override;
  end;

  TDataClass = class(TPanel)
   private
     TitleEl : TLabel;
     Box     : TScrollBox;
     _First  : TDataItem;
     _Last   : TDataItem;
     _Active : TDataItem;
     _xml    : XML.TElement;
     function ReadTitle : String; inline;
     procedure WriteTitle(v : String); inline;
     function GetItem(i : Cardinal) : TDataItem;
     procedure SetXML(el : XML.TElement);
     procedure NewItem(Sender : TObject);
   public
     AddBut  : TButton;
     tagName : ShortString;
     onCreateNode : function(tagName : ShortString) : TDataItem of object;
     constructor Create(_owner : TComponent);
     property Title : String read ReadTitle write WriteTitle;
     property Items[i : Cardinal] : TDataItem read GetItem;
     property Last : TDataItem read _Last;
     property First : TDataItem read _First;
     property XMLElement : XML.TElement read _xml write SetXML;
     destructor Destroy; override;
  end;

implementation

constructor TDataItem.Create;
begin
 inherited Create(nil);
 _Prev := nil;
 _Next := nil;
 _Parent := nil;
 TitleEl := TLabel.Create(self);
 TitleEl.Parent := self;
 TitleEl.AnchorParallel(akLeft, 4, self);
 TitleEl.AnchorParallel(akTop, 2, self);
 ButRem := TButton.Create(self);
 ButRem.Parent := self;
 ButRem.Caption := '-';
 ButRem.Width := ButRem.Height;
 ButRem.Anchors := [akTop, akBottom, akRight];
 ButRem.AnchorParallel(akTop, 2, self);
 ButRem.AnchorParallel(akRight, 20, self);
 ButRem.AnchorParallel(akBottom, 2, self);
 ButRem.OnClick := @RemoveClick;
 ButMod := TButton.Create(self);
 ButMod.Parent := self;
 ButMod.Caption := '...';
 ButMod.Width := ButMod.Height;
 ButMod.Anchors := [akTop, akBottom, akRight];
 ButMod.AnchorParallel(akTop, 2, self);
 ButMod.AnchorToNeighbour(akRight, 2, ButRem);
 ButMod.AnchorParallel(akBottom, 2, self);
 ButMod.OnClick := @Edit;
 self.Height := TitleEl.Height + 6;
 self.OnClick := @DoClick;
end;

procedure TDataItem.DoClick(Sender : TObject);
begin
 Activate;
end;

procedure TDataItem.Edit(Sender : TObject);
begin
 // Platzhalter - soll überschrieben werden!
end;

procedure TDataItem.Activate;
begin
 if (Parent = nil) or (Parent._Active = self) then exit;
 if (Parent._Active <> nil) then Parent._Active.Deactivate;
 self.Color := clBlue;
 Parent._Active := self;
end;

procedure TDataItem.Deactivate;
begin
 self.Color := clDefault;
 Parent._Active := nil;
end;

function TDataItem.ReadTitle : String; inline;
begin
 Result := TitleEl.Caption;
end;

procedure TDataItem.WriteTitle(v : String); inline;
begin
 TitleEl.Caption := v;
end;

procedure TDataItem.WriteParent(P : TDataClass);
begin
 _Parent := P;
 if (P <> nil) then begin
  SetParent(P.Box);
  Visible := true;
  AnchorParallel(akLeft, 0, P.Box);
  AnchorParallel(akRight, 0, P.Box);
 end else Visible := false;
end;

procedure TDataItem.Remove;
begin
 if (Prev <> nil) then Prev._Next := Next
 else if (Parent <> nil) then Parent._First := Next;
 if (Next <> nil) then begin Next._Prev := Prev; if (Prev = nil) then Next.AnchorParallel(akTop, 0, Parent) else Next.AnchorToNeighbour(akTop, 0, Prev); end
 else if (Parent <> nil) then Parent._Last := Prev;
 _Prev := nil;
 _Next := nil;
 if (Parent <> nil) and (Parent._Active = self) then Parent._Active := nil;
 Parent := nil;
end;

procedure TDataItem.RemoveClick(Sender : TObject);
begin
 if (MessageDlg('Löschen', 'Sind Sie sicher, dass Sie die diesen Datensatz löschen möchten?', mtConfirmation, [mbYes, mbNo], 0) = mrNo) then exit;
 Remove;
 Application.ReleaseComponent(self);
end;

procedure TDataItem.InsertAfter(P : TDataItem);
begin
 Remove;
 Parent := P.Parent;
 _Prev := P;
 _Next := P.Next;
 AnchorToNeighbour(akTop, 0, Prev);
 P._Next := self;
 if (Next <> nil) then begin Next._Prev := self; Next.AnchorToNeighbour(akTop, 0, self); end
 else Parent._Last := self;
end;

procedure TDataItem.InsertFirst(P : TDataClass);
begin
 Remove;
 Parent := P;
 _Next := Parent._First;
 AnchorParallel(akTop, 0, Parent.Box);
 Parent._First := self;
 if (Next <> nil) then begin Next._Prev := self; Next.AnchorToNeighbour(akTop, 0, self); end;
 if (Parent._Last = nil) then Parent._Last := self;
end;

destructor TDataItem.Destroy;
begin
  Remove;
  inherited;
end;




constructor TDataClass.Create(_owner : TComponent);
begin
  _First := nil;
  _Last := nil;
  _Active := nil;
  inherited;
  tagName := 'DATACLASS';
  TitleEl := TLabel.Create(self);
  TitleEl.Parent := self;
  TitleEl.AnchorParallel(akLeft, 2, self);
  TitleEl.AnchorParallel(akTop, 2, self);
  AddBut := TButton.Create(self);
  AddBut.Parent := self;
  AddBut.Caption := '+';
  AddBut.AutoSize := true;
  AddBut.AnchorParallel(akLeft, 2, self);
  AddBut.AnchorParallel(akBottom, 2, self);
  AddBut.Anchors := [akLeft, akBottom];
  AddBut.OnClick := @NewItem;
  AddBut.Enabled := false;
  Box := TScrollBox.Create(self);
  Box.Parent := self;
  Box.AnchorParallel(akLeft, 2, self);
  Box.AnchorParallel(akRight, 2, self);
  Box.AnchorToNeighbour(akTop, 2, TitleEl);
  Box.AnchorToNeighbour(akBottom, 2, AddBut);
end;

procedure TDataClass.NewItem(Sender : TObject);
var It : TDataItem;
begin
 It := onCreateNode('');
 if (It = nil) then exit;
 if (First = nil) then It.insertFirst(self) else It.insertAfter(Last);
end;

function TDataClass.ReadTitle : String; inline;
begin
 Result := TitleEl.Caption;
end;

procedure TDataClass.WriteTitle(v : String); inline;
begin
 TitleEl.Caption := v;
end;

function TDataClass.GetItem(i : Cardinal) : TDataItem;
begin
 Result := _First;
 while (i > 0) and (Result <> nil) do begin
  Result := Result.Next;
  i := i - 1;
 end;
end;

procedure TDataClass.SetXML(el : XML.TElement);
var
  i  : integer;
  It : TDataItem;
begin
 if (_xml <> nil) then begin
  _xml.clearNodes;
  It := First;
  while (It <> nil) do begin
   _xml.appendChild(It.DataToXML);
   It := It.Next;
  end;
 end;
 while (First <> nil) do First.Free;
 if (el = nil) then begin
  _xml := nil;
  AddBut.Enabled := false;
  exit;
 end else if (_xml = el) then exit;
 _xml := el;
 AddBut.Enabled := true;
 for i := _xml.childNodes.count - 1 downto 0 do if (_xml.childNodes[i].nodeType = ELEMENT_NODE) then begin
  It := onCreateNode(XML.TElement(_xml.childNodes[i]).tagName);
  It.XMLToData(XML.TElement(_xml.childNodes[i]));
  It.insertFirst(self);
 end;
end;

destructor TDataClass.Destroy;
begin
 while (First <> nil) do First.Free;
 inherited;
end;

end.

