unit Unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, StdCtrls,
  ExtCtrls, PairSplitter, Buttons, ComCtrls, EditBtn, Spin, Unit2,
  StrTransformErr, Graph2D, PolynomUtil, DataClasses, xmlUtil, Math;

type
 TVehicle = record
  m : Single;                             // Masse / Kg
  A : Single;                             // Querschnittsfläche / m^2
  Cw : function (v : Single) : Single;    // Luftwiderstandsbeiwert
  Froll : function (v : Single) : Single; // Rollreibung m/s → N
  h : function (F, v, a : Single) : Single;  // Motor und Getriebewirkungsgrad m/s, m/s^2 → W
 end;
 PVehicle = ^TVehicle;

 TFormGeschwMessung = class;

 TGeschwindigkeitsMessung = class(TDataItem)
  type
   TMeasurePoint = record
     Time     : Single;   // Sekunden nach Start
     Velocity : Single;   // m/s
     Power    : Single;   // w
     Name     : String;   // Name des Wegpunktes
     Way      : Single;   // Weg in m von 1. Wegpunkt
     Alt      : Single;   // Höhe des Wegpunktes über NN
   end;
   Poly = specialize Polynom<Extended>;
   PolyS = specialize Polynom<Single>;
   TEffectivePower = class(Poly.IFunction)
     private
      M : TGeschwindigkeitsmessung;
     public
      constructor Create(_M : TGeschwindigkeitsmessung);
      function calc(x : Extended) : Extended; override;
   end;

  private
   Start : TDateTime;
   Data  : Array of TMeasurePoint;
   Vehicle : PVehicle;
   Form  : TFormGeschwMessung;
   function ReadPoints(i : Cardinal) : TMeasurePoint; inline;
   procedure WritePoints(i : Cardinal; P : TMeasurePoint); inline;
   function ReadCount : Cardinal; inline;
   procedure WriteCount(c : Cardinal); inline;
   function ToStrings : TStrings;
   procedure FromStrings(St : TStrings);
   procedure updateInfo;
   procedure ClearFunctions;

  protected
   procedure CalcF;

  public
   Year         : Word;
   Month        : Word;
   Day          : Word;
   DayOfWeek    : Word;
   Hour         : Word;
   Minute       : Word;
   Second       : Word;
   mSecond      : Word;
   AdjustSpline : Extended;
   AdjustPoints : Extended;
   Fw  : Poly.IFunction;
   Fv  : Poly.IFunction;
   Fa  : Poly.IFunction;
   Fh  : Poly.IFunction;
   Fst : Poly.IFunction;
   Fpm : Poly.IFunction;
   Fpe : TEffectivePower;
   Wmess : Single;
   Wrech : Single;
   reverse : Boolean;
   constructor Create;
   procedure PackDate;
   procedure UnpackDate;
   property Points[i : Cardinal] : TMeasurePoint read ReadPoints write WritePoints; default;
   property Count : Cardinal read ReadCount write WriteCount;
   procedure Edit(Sender : TObject); override;
   function DataToXML : XML.TElement; override;
   procedure XMLToData(_Data : XML.TElement); override;
   destructor Destroy; override;
 end;

 { TFormGeschwMessung }

 TFormGeschwMessung = class(TForm)
   Button1: TButton;
   Button2: TButton;
   Button3: TButton;
   Button5: TButton;
   Button6: TButton;
   CheckBox1: TCheckBox;
   DateEdit1: TDateEdit;
   Edit1: TEdit;
   Label1: TLabel;
   Label2: TLabel;
   Label3: TLabel;
   Label4: TLabel;
   Label5: TLabel;
   Label6: TLabel;
   LabelWrech: TLabel;
   LabelWmess: TLabel;
   LabelMesspunkt: TLabel;
   Memo1: TMemo;
   PairSplitter1: TPairSplitter;
   PairSplitter2: TPairSplitter;
   PairSplitterSide1: TPairSplitterSide;
   PairSplitterSide2: TPairSplitterSide;
   GraphW: TGraph2D;
   GraphV: TGraph2D;
   PairSplitterSide3: TPairSplitterSide;
   PairSplitterSide4: TPairSplitterSide;
   Panel1: TPanel;
   SpinEdit1: TSpinEdit;
   SpinEdit2: TSpinEdit;
   TrackBar1: TTrackBar;
   TrackBar2: TTrackBar;
   procedure Add(Sender: TObject; var Key: Word; Shift: TShiftState);
   procedure AdjustViewRect(Sender: TObject);
   procedure DateChanged(Sender: TObject);
   procedure FormCreate(Sender: TObject);
   procedure FormResize(Sender: TObject);
   procedure KompensateChanged(Sender: TObject);
   procedure MessungChanged(Sender: TObject);
   procedure ReachPos(Sender: TObject);
   procedure ReverseChange(Sender: TObject);
   procedure Start(Sender: TObject);
   procedure Stopp(Sender: TObject);
   procedure OpenEnergyConsumption(Sender: TObject);
   procedure ApplyChanges(Sender: TObject);
   constructor Create(M : TGeschwindigkeitsMessung);
   destructor Destroy; override;

  private
   Pos     : integer;
   Messung : TGeschwindigkeitsMessung;
   Pmul    : Double;
   DateChange : Boolean;
   procedure RenderPoints(Graph : TGraph2D);
   procedure RenderGraphs(Graph : TGraph2D);
   procedure CheckViewRect;
   procedure AddValue(v : TGeschwindigkeitsMessung.TMeasurePoint);
 end;


implementation

function _Cw (v : Single) : Single;
begin
 Result := 0.31; // Quelle http://rc.opelgt.org/indexcw.php
end;

function _Froll (v : Single) : Single;
begin
 Result := 100/(20 + v) + 30;
end;

function _h (f, v, a : Single) : Single;
var s : Single; // Wirkungsgrad
begin
 if (f < 0) then f := 0; // Bremsleistung ist verloren
 s := (a / (2 + a)) * (0.2 + 0.8 * 11 / (11 + v)) * 0.3 + 0.8 * 60 / (60 + v*v);
 s := 0.35*(1-s) + 0.04*s;
 Result := 1/s * f * v + 11000{Standleistung};
end;

var
  Schrottkarre : TVehicle = (m : 1400; A : 2.26; Cw : @_Cw; Froll : @_Froll; h : @_h);





type
 Poly = TGeschwindigkeitsmessung.Poly;
 PolyS = TGeschwindigkeitsmessung.PolyS;
 Precision = Extended;

const
  Tab : String = Char(9);

{$R *.lfm}



constructor TGeschwindigkeitsMessung.Create;
begin
 inherited;
 SetLength(Data, 0);
 Form := nil;
 Fw := nil;
 Fv := nil;
 Fa := nil;
 Fpm := nil;
 Fpe := nil;
 Fh := nil;
 Fst := nil;
 Wmess := 0;
 Vehicle := @Schrottkarre;
 AdjustSpline := 0.1;
 AdjustPoints := 0.1;
 updateInfo;
end;

procedure TGeschwindigkeitsMessung.PackDate;
begin
 Start := EncodeDate(Year, Month, Day) + EncodeTime(Hour, Minute, Second, mSecond);
 DayOfWeek := sysutils.DayOfWeek(Start);
 Title := sysutils.FormatDateTime('YYYY.MM.DD hh:nn', Start);
 if (Form <> nil) then begin
  Form.Caption := Title;
 end;
end;

procedure TGeschwindigkeitsMessung.UnpackDate;
begin
 sysutils.DecodeDateFully(Start, Year, Month, Day, DayOfWeek);
 sysutils.DecodeTime(Start, Hour, Minute, Second, mSecond);
 Title := sysutils.FormatDateTime('YYYY.MM.DD hh:nn', Start);
 if (Form <> nil) then begin
  Form.Caption := Title;
  Form.DateEdit1.Date := Start;
  Form.SpinEdit1.Value := Hour;
  Form.SpinEdit2.Value := Minute;
  Form.DateChange := false;
 end;
end;

procedure TGeschwindigkeitsMessung.updateInfo;
var
  h, m, s, ms : Word;
begin
 if (Length(Data) = 0) then begin
  Hour := 0;
  DayOfWeek := 0;
  if (Form <> nil) then Form.Caption := Title;
  exit;
 end;
 UnpackDate;
end;

function TGeschwindigkeitsMessung.ReadPoints(i : Cardinal) : TMeasurePoint; inline;
begin
 Result := Data[i];
end;

procedure TGeschwindigkeitsMessung.WritePoints(i : Cardinal; P : TMeasurePoint); inline;
begin
 Data[i] := P;
end;

function TGeschwindigkeitsMessung.ReadCount : Cardinal; inline;
begin
 Result := Length(Data);
end;

procedure TGeschwindigkeitsMessung.WriteCount(c : Cardinal); inline;
begin
 SetLength(Data, c);
end;

procedure TGeschwindigkeitsMessung.Edit(Sender : TObject);
var L : TStrings;
begin
 if (Form = nil) then Form := TFormGeschwMessung.Create(self);
 Form.Show;
 L := ToStrings;
 Form.Memo1.Lines.Assign(L);
 L.Free;
 CalcF;
 Form.CheckViewRect;
 updateInfo;
 Form.Button5.Enabled := false;
 Form.CheckBox1.Checked := self.reverse;
end;

procedure TGeschwindigkeitsMessung.FromStrings(St : TStrings);
var
  i,j,k,c : integer;
  SelPos : Array [0..5] of integer;
  P : TGeschwindigkeitsMessung.TMeasurePoint;
  s : String;
  M : Array of TGeschwindigkeitsMessung.TMeasurePoint;
begin
 c := 0;
 SetLength(M, 0);
 if (Form <> nil) and (Form.Memo1.Lines <> St) then Form.Memo1.Lines := St;
 for i := 0 to St.Count - 1 do begin
  if (St[i] = '') then continue;
  P.Time := -1;
  P.Velocity := PolyS.NaN;
  P.Power := PolyS.NaN;
  P.Name := '';
  P.Way := PolyS.NaN;
  P.Alt := PolyS.NaN;
  k := 0;
  s := '';
  for j := 1 to Length(St[i]) do begin
   if (St[i][j] <> Tab) then begin
    s := s + St[i][j];
    if (j <> Length(St[i])) then continue;
   end;
   case k of
    0 : P.Time := StrToFloatErr(s, -1);
    1 : P.Velocity := StrToFloatErr(s, PolyS.NaN);
    2 : P.Power := StrToFloatErr(s, PolyS.NaN);
    3 : P.Name := s;
    4 : P.Way := StrToFloatErr(s, PolyS.NaN);
    5 : P.Alt := StrToFloatErr(s, PolyS.NaN);
   end;
   if (k < Length(SelPos)) then SelPos[k] := j;
   s := '';
   k := k + 1;
  end;
  if (P.Time < 0) and (Form <> nil) then begin
   Form.memo1.selstart := c;
   Form.Memo1.SelLength := SelPos[0]-1;
   Form.Memo1.SetFocus;
   if (MessageDlg('Ungültiger Messpunkt', 'Die Zeile ' + IntToStr(i) + '("' + St[i] + '") hat keinen gültigen Zeitstempel. Soll diese Zeile übergangen werden?', mtConfirmation, [mbIgnore, mbAbort],0) = mrIgnore) then continue else exit;
  end else if (P.Time < 0) then continue;
  c := c + Length(St[i]) + 2; // Absatzgröße hier!
  k := Length(M);
  SetLength(M, k+1);
  M[k] := P;
 end;
 Data := M;
 ClearFunctions;
 if (Form <> nil) then begin
  Form.CheckViewRect;
  CalcF;
  Form.Button5.Enabled := false;
 end;
 updateInfo;
end;

function TGeschwindigkeitsMessung.ToStrings : TStrings;
var
  i : integer;
  s : String;
begin
 Result := TStringList.Create;
 for i := 0 to Length(Data) - 1 do begin
  s := FloatToStr(Data[i].Time) + Tab;
  if (not PolyS.IsNaN(Data[i].Velocity)) then s := s + FloatToStr(Data[i].Velocity);
  s := s + Tab;
  if (not PolyS.IsNaN(Data[i].Power)) then s := s + FloatToStr(Data[i].Power);
  s := s + Tab + Data[i].Name + Tab;
  if (not PolyS.IsNaN(Data[i].Way)) then s := s + FloatToStr(Data[i].Way);
  s := s + Tab;
  if (not PolyS.IsNaN(Data[i].Alt)) then s := s + FloatToStr(Data[i].Alt);
  Result.add(s);
 end;
end;

function TGeschwindigkeitsMessung.DataToXML : XML.TElement;
begin
 Result := XML.TElement.Create;
 Result.tagName := 'MESSUNG';
 Result.setAttribute('title', title);
 Result.setAttribute('start', FloatToStr(Start));
 if (reverse) then Result.setAttribute('reverse', 'true') else Result.setAttribute('reverse', 'false');
 Result.appendChild(XML.TText.Create(ToStrings));
end;

procedure TGeschwindigkeitsMessung.XMLToData(_Data : XML.TElement);
var S : TStrings;
begin
 SetLength(Data, 0);
 if (_Data.tagName <> 'MESSUNG') then exit;
 title := _Data.GetAttribute('title');
 Start := StrToFloatErr(_Data.GetAttribute('start'), 0);
 reverse := _Data.GetAttribute('reverse') = 'true';
 S := _Data.getTextContent(true);
 FromStrings(S);
 S.Free;
end;

constructor TGeschwindigkeitsMessung.TEffectivePower.Create(_M : TGeschwindigkeitsmessung);
begin
 inherited Create;
 M := _M;
end;

function TGeschwindigkeitsMessung.TEffectivePower.Calc(x : Extended) : Extended;
var
  F : Extended;
  v : Extended;
  a, a_ : Extended;
  s : Extended;
begin
 with (M) do begin
  v := Fv.Calc(x); if (Poly.IsNaN(v)) then v := 0;
  a := 0;
  a_ := Fa.Calc(x-0.5)  * 0.0625; if (not Poly.IsNaN(a_)) then a := a + a_;
  a_ := Fa.Calc(x-0.25) * 0.1875; if (not Poly.IsNaN(a_)) then a := a + a_;
  a_ := Fa.Calc(x)      * 0.5;    if (not Poly.IsNaN(a_)) then a := a + a_;
  a_ := Fa.Calc(x+0.25) * 0.1875; if (not Poly.IsNaN(a_)) then a := a + a_;
  a_ := Fa.Calc(x+0.5)  * 0.0625; if (not Poly.IsNaN(a_)) then a := a + a_;
  s := (Fw.Calc(x+0.5)-Fw.Calc(x-0.5));
  if (Poly.IsNaN(s)) then s := 0 else if (s <> 0) then s := (Fh.Calc(x+0.5)-Fh.Calc(x-0.5)) / s;   // s : dh/dw
  F := Vehicle^.m{kg} * a{m/s^2};  // Beschleunigung
  if (s <> 0) then F := F + Vehicle^.m{kg} * 9.81{m/s^3} * s; // Hangabtrieb
  F := F + Vehicle^.Froll(v){N}; // Rollreibung
  F := F + Vehicle^.A{m^2} * Vehicle^.Cw(v) * 1/2 * 1.293{kg/m^3} * v*v{m/s}; // Luftwiderstand
  Result := Vehicle^.h(F, v, a); // Motorwirkungsgrad
 end;
end;

procedure TGeschwindigkeitsmessung.ClearFunctions;
begin
 Fw.Free; Fw := nil;
 Fv.Free; Fv := nil;
 Fa.Free; Fa := nil;
 Fpm.Free; Fpm := nil;
 Fpe.Free; Fpe := nil;
 Fh.Free; Fh := nil;
 Fst.Free; Fst := nil;
 Wmess := 0;
end;

procedure TGeschwindigkeitsMessung.CalcF;
var
  IPvw, IPh, IPpm : Array of Poly.TInterpolatePoint;
  nvw,  nh,  npm  : Integer;
  mint, maxt : Single;
  i,j : integer;
  f : Poly.TPolySpline;
begin
 //if (not self.Button1.Enabled) then exit;
 ClearFunctions;
 SetLength(IPvw, Count * 2);
 SetLength(IPh, Route.Count);
 SetLength(IPpm, Count);
 nvw := -1;
 nh := -1;
 npm := -1;
 mint := 100000000000000;
 maxt := -100000000000000;
 for i := Count - 1 downto 0 do begin
  if (self[i].Time < mint) then mint := self[i].Time;
  if (self[i].Time > maxt) then maxt := self[i].Time;
  if (not PolyS.IsNaN(self[i].Way)) then begin
   nvw := nvw + 1;
   IPvw[nvw].x := self[i].Time;
   IPvw[nvw].y := self[i].Way;
   IPvw[nvw].derivation := 0;
  end;
  if (not PolyS.IsNaN(self[i].Velocity)) then begin
   nvw := nvw + 1;
   IPvw[nvw].x := self[i].Time;
   IPvw[nvw].y := self[i].Velocity;
   IPvw[nvw].derivation := 1;
  end;
  if (not PolyS.IsNaN(self[i].Alt)) then begin
   nh := nh + 1;
   IPh[nh].x := self[i].Time;
   IPh[nh].y := self[i].Alt;
   IPh[nh].derivation := 0;
  end;
  if (not PolyS.IsNaN(self[i].Power)) then begin
   npm := npm + 1;
   IPpm[npm].x := self[i].Time;
   IPpm[npm].y := self[i].Power;
   IPpm[npm].derivation := 0;
  end;
 end;
 SetLength(IPvw, nvw+1);
 SetLength(IPpm, npm+1);
 Fw := Poly.AdjustedSplinesOnFirstDerivation(IPvw, AdjustSpline, AdjustPoints);
 Fv := Fw.Derivate;
 Fa := Fv.Derivate;
 if (self.reverse) then begin for i := 0 to Route.Count - 1 do if (Route.Reverse[i].Name = '') then begin
  nh := nh + 1;
  IPh[nh].derivation := 0;
  IPh[nh].x := Fw.CalcInverse(Route.Reverse[i].Way, mint, maxt, Fv);
  IPh[nh].y := Route.Reverse[i].Alt;
 end end else for i := 0 to Route.Count - 1 do if (Route[i].Name = '') then begin
  nh := nh + 1;
  IPh[nh].derivation := 0;
  IPh[nh].x := Fw.CalcInverse(Route[i].Way, mint, maxt, Fv);
  IPh[nh].y := Route[i].Alt;
 end;
 j := -1;
 for i := 0 to nh do if (not Poly.IsNaN(IPh[i].x)) then begin // Falls Höhenpunkte nicht interpoliert werden konnten, diese entfernen.
  j := j+1;
  IPh[j] := IPh[i];
 end;
 SetLength(IPh, j+1);
 Fh := Poly.KubicSplines(IPh);
 Fst := Fh.Derivate;
 Fpm := Poly.KubicSplines(IPpm);
 Fpe := TEffectivePower.Create(self);
 // Energie berechnen
 f := Poly.TPolySpline(Fpm).Integrate(0);
 if (f.Count > 0) then Wmess := f[f.Count-1].fkt.calc(f[f.Count-1].max) - f[0].fkt.calc(f[0].min);
 f.Free;
 Wrech := Fpe.TrapezIntegral(mint, maxt, 100000);
 if (Form <> nil) then begin
  Form.LabelWmess.Caption := FloatToStr(Wmess/1000000) + 'MJ';
  Form.LabelWrech.Caption := FloatToStr(Wrech/1000000) + 'MJ';
  Form.GraphW.DoPaint(nil);
  Form.GraphV.DoPaint(nil);
 end;
end;

destructor TGeschwindigkeitsMessung.Destroy;
begin
 Form.Free;
 SetLength(Data, 0);
 ClearFunctions;
 inherited;
end;





constructor TFormGeschwMessung.Create(M : TGeschwindigkeitsMessung);
begin
 inherited Create(nil);
 Messung := M;
 Messung.Form := self;
 Caption := Messung.Title;
 DateChange := false;
end;

procedure TFormGeschwMessung.FormCreate(Sender: TObject);
begin
 GraphW := TGraph2D.Create(self);
 GraphW.Parent := PairSplitterSide3;
 GraphW.AnchorParallel(akTop, 0, PairSplitterSide3);
 GraphW.AnchorParallel(akLeft, 0, PairSplitterSide3);
 GraphW.AnchorParallel(akRight, 0, PairSplitterSide3);
 GraphW.AnchorParallel(akBottom, 0, PairSplitterSide3);
 SetLength(GraphW.RenderFkts, 2);
 GraphW.RenderFkts[0] := TMethod(@RenderPoints);
 GraphW.RenderFkts[1] := TMethod(@RenderGraphs);
 GraphW.Axis[0].Name := 't/s';
 GraphW.Axis[1].Name := 'l/m';
 GraphV := TGraph2D.Create(self);
 GraphV.Parent := PairSplitterSide4;
 GraphV.AnchorParallel(akTop, 0, PairSplitterSide4);
 GraphV.AnchorParallel(akLeft, 0, PairSplitterSide4);
 GraphV.AnchorParallel(akRight, 0, PairSplitterSide4);
 GraphV.AnchorParallel(akBottom, 0, PairSplitterSide4);
 GraphV.Axis[0].Name := 't/s';
 GraphV.Axis[1].Name := 'v/(m/s)';
 GraphV.Axis[1].Position := - TGraph2D.MaxRange;
 SetLength(GraphV.Axis, 3);
 GraphV.Axis[2].Init;
 GraphV.Axis[2].Vertical := true;
 GraphV.Axis[2].Name := 'P/KW';
 GraphV.Axis[2].Color := clBlue;
 GraphV.Axis[2].Position := TGraph2D.MaxRange;
 Pmul := GraphV.Axis[2].Multiplicator;
 SetLength(GraphV.RenderFkts, 2);
 GraphV.RenderFkts[0] := TMethod(@RenderPoints);
 GraphV.RenderFkts[1] := TMethod(@RenderGraphs);
end;

procedure TFormGeschwMessung.Start(Sender: TObject);
begin
 if (Route.Count = 0) then begin
  ShowMessage('Messung kann nicht gestartet werden, da keine Route definiert wurde!');
  exit;
 end;
 Memo1.Lines.Clear;
 SetLength(Messung.Data, 0);
 Messung.Start := 0;
 Pos := 0;
 Button1.enabled := false;
 Button2.enabled := true;
 Button3.enabled := true;
 Edit1.enabled := true;
 if (CheckBox1.Checked) then LabelMesspunkt.Caption := Route.Reverse[Pos].Name
 else LabelMesspunkt.Caption := Route[Pos].Name;
 Memo1.ReadOnly := true;
 Button5.Enabled := false;
 CheckBox1.Enabled := false;
end;

procedure TFormGeschwMessung.ReachPos(Sender: TObject);
var v : TGeschwindigkeitsMessung.TMeasurePoint;
begin
 if (Pos >= Route.Count) then exit;
 v.Velocity := StrToFloatErr(Edit1.Text, PolyS.NaN);
 if (not PolyS.IsNaN(v.Velocity)) then v.Velocity := v.Velocity * 0.2777777777777777;
 if (Messung.Start = 0) then begin Messung.Start := Now; v.Time := 0; end
 else v.Time := (Now-Messung.Start)*100000;
 if (CheckBox1.Checked) then begin
  v.Name := Route.Reverse[Pos].Name;
  v.Way := Route.Reverse[Pos].Way;
  v.Alt := Route.Reverse[Pos].Alt;
 end else begin
  v.Name := Route[Pos].Name;
  v.Way := Route[Pos].Way;
  v.Alt := Route[Pos].Alt;
 end;
 v.Power := PolyS.NaN;
 AddValue(v);
 Edit1.Text := '';
 Edit1.SetFocus;
 repeat Pos := Pos + 1 until (Pos >= Route.Count) or (Route[Pos].Name <> '');
 if (Pos >= Route.Count) then Stopp(Sender)
 else if (CheckBox1.Checked) then LabelMesspunkt.Caption := Route.Reverse[Pos].Name
 else LabelMesspunkt.Caption := Route[Pos].Name;
end;

procedure TFormGeschwMessung.ReverseChange(Sender: TObject);
begin
 Messung.reverse := CheckBox1.Checked;
end;

procedure TFormGeschwMessung.Add(Sender: TObject; var Key: Word; Shift: TShiftState);
var
  v : TGeschwindigkeitsMessung.TMeasurePoint;
  t : Double;
begin
 if (Key <> 13) or (Messung.Count = 0) then exit;
 v.Velocity := StrToFloatErr(Edit1.Text, PolyS.NaN);
 if (PolyS.IsNaN(v.Velocity)) then begin
  Edit1.SelectAll;
  exit;
 end;
 v.Velocity := v.Velocity * 0.2777777777777777;
 v.Power := PolyS.NaN;
 if (Messung.Start = 0) then begin Messung.Start := Now; v.Time := 0; end
 else v.Time := (Now-Messung.Start)*100000;
 v.Name := '';
 v.Way := PolyS.NaN;
 v.Alt := PolyS.NaN;
 AddValue(v);
 Edit1.Text := '';
end;

procedure TFormGeschwMessung.AdjustViewRect(Sender: TObject);
begin
 CheckViewRect;
end;

procedure TFormGeschwMessung.DateChanged(Sender: TObject);
begin
 DateChange := true;
 Button5.Enabled := true;
end;

procedure TFormGeschwMessung.Stopp(Sender: TObject);
var S : TStrings;
begin
  Button1.enabled := true;
  Button2.enabled := false;
  Button3.enabled := false;
  Edit1.enabled := false;
  CheckBox1.Enabled := true;
  LabelMesspunkt.Caption := '';
  Edit1.Text := '';
  S := Messung.ToStrings;
  Memo1.Lines.Assign(S);
  S.Free;
  Memo1.ReadOnly := false;
  Messung.CalcF;
end;

procedure TFormGeschwMessung.OpenEnergyConsumption(Sender: TObject);
var
  p : Poly.TPolynom;
  m : Poly.TMatrix;
begin
  p := Poly.TPolynom.Create;
  p[0] := 1;
  p[1] := 3;
  p[2] := 2;
  ShowMessage(p.ToString);
  m.Init(4, 3, false);
  m[3,2] := 5.32;
  ShowMessage(FloatToStr(m[3,2]));
end;

procedure TFormGeschwMessung.ApplyChanges(Sender: TObject);
begin
 if (DateChange) then begin
  DateChange := false;
  sysutils.DecodeDate(DateEdit1.Date, Messung.Year, Messung.Month, Messung.Day);
  Messung.Hour := SpinEdit1.Value;
  Messung.Minute := SpinEdit2.Value;
  Messung.PackDate;
 end;
 Messung.FromStrings(Memo1.Lines);
end;

procedure TFormGeschwMessung.FormResize(Sender: TObject);
begin
  if (self.Width < 900) then self.Width := 900;
end;

procedure TFormGeschwMessung.KompensateChanged(Sender: TObject);
begin
  Messung.AdjustSpline := Trackbar1.Position / Trackbar1.Max;
  Messung.AdjustPoints := Trackbar2.Position / Trackbar2.Max;
  Messung.CalcF;
end;

procedure TFormGeschwMessung.MessungChanged(Sender: TObject);
begin
  if (Button1.enabled) then Button5.Enabled := true;
end;

procedure TFormGeschwMessung.RenderPoints(Graph : TGraph2D);
var
  i, x, y : integer;
  P : Array [0..2] of TPoint;
begin
  if (Messung.Count = 0) then exit;
  for i := 0 to Messung.Count-1 do begin
   x := Graph.XPosToScreen(Messung[i].Time);
   if (Graph = self.GraphW) then begin
    if (not PolyS.IsNaN(Messung[i].Way)) then begin
     Graph.SetColor(1);
     y := Graph.YPosToScreen(Messung[i].Way);
     Graph.Canvas.FillRect(x-4, y-4, x+4, y+4);
    end;
    if (not PolyS.IsNaN(Messung[i].Alt)) then begin
     Graph.SetColor(1);
     y := Graph.YPosToScreen(Messung[i].Alt);
     Graph.Canvas.FillRect(x-4, y-4, x+4, y+4);
    end;
   end else if (Graph = self.GraphV) then begin
    if (not PolyS.IsNaN(Messung[i].Velocity)) then begin
     Graph.SetColor(1);
     y := Graph.YPosToScreen(Messung[i].Velocity);
     P[0].x := x-4;  P[0].y := y+4;
     P[1].x := x+4;  P[1].y := y+4;
     P[2].x := x;  P[2].y := y-4;
     Graph.Canvas.Polygon(P);
    end;
    if (not PolyS.IsNaN(Messung[i].Power)) then begin
     Graph.SetColor(2);
     y := Graph.YPosToScreen(Messung[i].Power/1000*Pmul);
     Graph.Canvas.Ellipse(x-4, y-4, x+4, y+4);
    end;
    if (not PolyS.IsNaN(Messung[i].Way)) then begin
     Graph.SetColor(1);
     Graph.Canvas.MoveTo(x, 0);
     Graph.Canvas.LineTo(x, Graph.height);
    end;
   end;
  end;
end;

procedure TFormGeschwMessung.RenderGraphs(Graph : TGraph2D);
procedure Render(P : Poly.IFunction; C : Integer; Style : Byte; mul : Double);
var
  f      : Precision;
  b      : Boolean;
  x, y   : integer;
  _x, _y : integer;
  i,j    : integer;
  w      : integer;
begin
  if (P = nil) then exit;
  Graph.SetColor(c);
  b := true;
  w := 0;
  for x := 0 to Graph.Canvas.Width - 1 do begin
   f := P.Calc(Graph.XScreenToPos(x));
   if (Poly.IsNaN(f)) then begin b := true; continue; end;
   y := Graph.YPosToScreen(f*mul);
   if (b) then begin Graph.Canvas.MoveTo(x, y); _x := x; _y := y; b := false; end
   else case Style of
    0 : Graph.Canvas.LineTo(x, y);
    1 : begin
     if (x = _x) then continue;
     for i := _x to x-1 do if (y > _y) then for j := Floor((i-_x)*(y-_y)/(x-_x) + _y) to Ceil((i+1-_x)*(y-_y)/(x-_x) + _y) do begin
      if ((w div 3) mod 2 = 0) then Graph.Canvas.Pixels[i, j] := Graph.Axis[c].Color;
      w := w + 1;
     end else for j := Ceil((i-_x)*(y-_y)/(x-_x) + _y) downto Floor((i+1-_x)*(y-_y)/(x-_x) + _y) do begin
      if ((w div 3) mod 2 = 0) then Graph.Canvas.Pixels[i, j] := Graph.Axis[c].Color;
      w := w + 1;
     end;
     _x := x;
     _y := y;
    end;
   end;
  end;
end;
begin
 if (Graph = self.GraphW) then begin
  Render(Messung.Fw, 1, 0, 1);
  Render(Messung.Fh, 1, 1, 1);
 end else begin
  Render(Messung.Fv, 1, 0, 1);
  Render(Messung.Fpm, 2, 0, Pmul/1000);
  Render(Messung.Fpe, 2, 1, Pmul/1000);
 end;
end;

procedure TFormGeschwMessung.CheckViewRect;
var
 i : integer;
 t : double;
 minX, maxX, minYw, maxYw, minYv, maxYv, maxYp : double;

function lt(A, B : Single) : Boolean;
begin
 if (PolyS.IsNaN(B)) then Result := true
 else if (PolyS.IsNaN(A)) then Result := false
 else Result := A < B;
end;
function gt(A, B : Single) : Boolean;
begin
 if (PolyS.IsNaN(B)) then Result := true
 else if (PolyS.IsNaN(A)) then Result := false
 else Result := A > B;
end;

begin
 minX := 0;
 maxX := 1;
 minYw := 0;
 maxYw := 1;
 minYv := 0;
 maxYv := 1;
 maxYp := 0.1;
 for i := 0 to Messung.Count - 1 do begin
  t := Messung[i].Time;
  if (lt(t, minX)) then minX := t;
  if (gt(t, maxX)) then maxX := t;
  if (lt(Messung[i].Way, minYw)) then minYw := Messung[i].Way;
  if (gt(Messung[i].Way, maxYw)) then maxYw := Messung[i].Way;
  if (lt(Messung[i].Velocity, minYv)) then minYv := Messung[i].Velocity;
  if (gt(Messung[i].Velocity, maxYv)) then maxYv := Messung[i].Velocity;
  if (gt(Messung[i].Power, maxYp)) then maxYp := Messung[i].Power;
  if (Messung.Fpe <> nil) and (gt(Messung.Fpe.Calc(Messung[i].Time), maxYp)) then maxYp := Messung.Fpe.Calc(Messung[i].Time);
 end;
 t := (maxX - minX)*0.05;
 maxX := maxX + t;
 minX := minX - t;
 t := (maxYw - minYw)*0.05;
 maxYw := maxYw + t;
 minYw := minYw - t;
 t := (maxYv - minYv)*0.05;
 maxYv := maxYv + t;
 minYv := minYv - t;
 Pmul := maxYv/maxYp*1000;
 self.GraphV.Axis[2].Multiplicator := Pmul;
 self.GraphW.SetViewArea(minX, minYw, maxX, maxYw);
 self.GraphW.DoPaint(nil);
 self.GraphV.SetViewArea(minX, minYv, maxX, maxYv);
 self.GraphV.DoPaint(nil);
end;

procedure TFormGeschwMessung.AddValue(v : TGeschwindigkeitsMessung.TMeasurePoint);
var
  l : integer;
  s : String;
begin
 l := Messung.Count;
 Messung.Count := l+1;
 Messung[l] := v;
 s := FloatToStr(v.Time) + Tab;
 if (not PolyS.IsNaN(v.Velocity)) then s := s + FloatToStr(v.Velocity);
 s := s + Tab;
 if (not PolyS.IsNaN(v.Power)) then s := s + FloatToStr(v.Power);
 s := s + Tab + v.Name + Tab;
 if (not PolyS.IsNaN(v.Way)) then s := s + FloatToStr(v.Way);
 s := s + Tab;
 if (not PolyS.IsNaN(v.Alt)) then s := s + FloatToStr(v.Alt);
 self.Memo1.Lines.Add(s);
 CheckViewRect;
 if (l = 0) then Messung.updateInfo;
end;

destructor TFormGeschwMessung.Destroy;
begin
 Messung.Form := nil;
 inherited;
end;

end.

